package NGCP::BulkProcessor::Projects::Migration::IPGallery::Dao::import::Lnp;
use strict;

## no critic

use NGCP::BulkProcessor::Projects::Migration::IPGallery::ProjectConnectorPool qw(
    get_import_db
    destroy_all_dbs
);

use NGCP::BulkProcessor::SqlProcessor qw(
    registertableinfo
    create_targettable
    checktableinfo
    copy_row

    insert_stmt

    process_table
);
use NGCP::BulkProcessor::SqlRecord qw();

use NGCP::BulkProcessor::Array qw(contains);

require Exporter;
our @ISA = qw(Exporter NGCP::BulkProcessor::SqlRecord);
our @EXPORT_OK = qw(
    create_table
    gettablename
    check_table
    getinsertstatement
    getupsertstatement

    findby_lrncode_portednumber
    list_lrncodes
    list_lrncodes_delta
    countby_lrncode_portednumber
    count_lrncodes

    update_delta
    findby_delta
    countby_delta

    $deleted_delta
    $updated_delta
    $added_delta

    process_records

    $IN_TYPE
);

my $tablename = 'lnp';
my $get_db = \&get_import_db;
#my $get_tablename = \&import_db_tableidentifier;


my $expected_fieldnames = [
    'ported_number',
    'type',
    'lrn_code',
    'delta',
];

# table creation:
my $primarykey_fieldnames = [ 'ported_number' ];
my $indexes = { $tablename . '_delta' => [ 'delta(7)' ]};
#my $fixtable_statements = [];

our $deleted_delta = 'DELETED';
our $updated_delta = 'UPDATED';
our $added_delta = 'ADDED';

our $IN_TYPE = 'In';

sub new {

    my $class = shift;
    my $self = NGCP::BulkProcessor::SqlRecord->new($class,$get_db,
                           $tablename,$expected_fieldnames,$indexes);

    copy_row($self,shift,$expected_fieldnames);

    return $self;

}

sub create_table {

    my ($truncate) = @_;

    my $db = &$get_db();

    registertableinfo($db,__PACKAGE__,$tablename,$expected_fieldnames,$indexes,$primarykey_fieldnames);
    return create_targettable($db,__PACKAGE__,$db,__PACKAGE__,$tablename,$truncate,0,undef);

}

sub findby_delta {

    my ($delta,$load_recursive) = @_;

    check_table();
    my $db = &$get_db();
    my $table = $db->tableidentifier($tablename);

    return [] unless defined $delta;

    my $rows = $db->db_get_all_arrayref(
        'SELECT * FROM ' .
            $table .
        ' WHERE ' .
            $db->columnidentifier('delta') . ' = ?'
    ,$delta);

    return buildrecords_fromrows($rows,$load_recursive);

}

sub findby_lrncode_portednumber {

    my ($lrncode,$portednumber,$load_recursive) = @_;

    check_table();
    my $db = &$get_db();
    my $table = $db->tableidentifier($tablename);

    my $rows = $db->db_get_all_arrayref(
        'SELECT * FROM ' .
            $table .
        ' WHERE ' .
            $db->columnidentifier('lrn_code') . ' = ? ' .
            ' AND ' . $db->columnidentifier('ported_number') . ' = ?'
    ,$lrncode,$portednumber);

    return buildrecords_fromrows($rows,$load_recursive)->[0];

}

sub list_lrncodes {

    my ($deltas) = @_;
    check_table();
    my $db = &$get_db();
    my $table = $db->tableidentifier($tablename);

    my $stmt = 'SELECT DISTINCT(' .
        $db->columnidentifier('lrn_code') . ') FROM ' . $table . ' WHERE 1=1';
    my @params = ();
    if (defined $deltas and 'HASH' eq ref $deltas) {
        foreach my $in (keys %$deltas) {
            my @values = (defined $deltas->{$in} and 'ARRAY' eq ref $deltas->{$in} ? @{$deltas->{$in}} : ($deltas->{$in}));
            $stmt .= ' AND ' . $db->columnidentifier('delta') . ' ' . $in . ' (' . substr(',?' x scalar @values,1) . ')';
            push(@params,@values);
        }
    } elsif (defined $deltas and length($deltas) > 0) {
        $stmt .= ' AND ' . $db->columnidentifier('delta') . ' = ?';
        push(@params,$deltas);
    }

    return $db->db_get_col($stmt,@params);

}

sub list_lrncodes_delta {

    check_table();
    my $db = &$get_db();
    my $table = $db->tableidentifier($tablename);

    my $stmt = 'SELECT ' . $db->columnidentifier('lrn_code') .
        ', ' . $db->columnidentifier('delta') . ' FROM ' . $table . ' GROUP BY ' .
        $db->columnidentifier('lrn_code') . ', ' . $db->columnidentifier('delta');
    my %lrn_codes = ();
    foreach my $lrn_code (@{$db->db_get_all_arrayref($stmt)}) {
        if (not exists $lrn_codes{$lrn_code->{'lrn_code'}}) {
            $lrn_codes{$lrn_code->{'lrn_code'}} = [];
        }
        push(@{$lrn_codes{$lrn_code->{'lrn_code'}}},$lrn_code->{'delta'});
    }
    my @result = ();
    foreach my $lrn_code (keys %lrn_codes) {
        my $row = { lrn_code => $lrn_code, delta => undef, };
        if (contains($updated_delta,$lrn_codes{$lrn_code})) {
            $row->{delta} = $updated_delta;
        } elsif (contains($added_delta,$lrn_codes{$lrn_code})) {
            if ((scalar @{$lrn_codes{$lrn_code}}) == 1) {
                $row->{delta} = $added_delta;
            } else {
                $row->{delta} = $updated_delta;
            }
        } elsif (contains($deleted_delta,$lrn_codes{$lrn_code})) {
            $row->{delta} = $deleted_delta;
        }
        push(@result,$row);
    }
    return \@result;

}

sub update_delta {

    my ($lrncode,$portednumber,$delta) = @_;

    check_table();
    my $db = &$get_db();
    my $table = $db->tableidentifier($tablename);

    my $stmt = 'UPDATE ' . $table . ' SET delta = ?';
    my @params = ();
    push(@params,$delta);
    if (defined $lrncode) {
        $stmt .= ' WHERE ' . $db->columnidentifier('lrn_code') . ' = ?';
        push(@params,$lrncode);
        if (defined $portednumber) {
            $stmt .= ' AND ' . $db->columnidentifier('ported_number') . ' = ?';
            push(@params,$portednumber);
        }
    }

    return $db->db_do($stmt,@params);

}

sub countby_lrncode_portednumber {

    my ($lrncode,$portednumber) = @_;

    check_table();
    my $db = &$get_db();
    my $table = $db->tableidentifier($tablename);

    my $stmt = 'SELECT COUNT(*) FROM ' . $table;
    my @params = ();
    if (defined $lrncode) {
        $stmt .= ' WHERE ' . $db->columnidentifier('lrn_code') . ' = ?';
        push(@params,$lrncode);
        if (defined $portednumber) {
            $stmt .= ' AND ' . $db->columnidentifier('ported_number') . ' = ?';
            push(@params,$portednumber);
        }
    }

    return $db->db_get_value($stmt,@params);

}

sub count_lrncodes {

    check_table();
    my $db = &$get_db();
    my $table = $db->tableidentifier($tablename);

    return $db->db_get_value('SELECT COUNT(DISTINCT ' .
        $db->columnidentifier('lrn_code') . ') FROM ' . $table);

}

sub countby_delta {

    my ($deltas) = @_;

    check_table();
    my $db = &$get_db();
    my $table = $db->tableidentifier($tablename);

    my $stmt = 'SELECT COUNT(*) FROM ' . $table . ' WHERE 1=1';
    my @params = ();
    if (defined $deltas and 'HASH' eq ref $deltas) {
        foreach my $in (keys %$deltas) {
            my @values = (defined $deltas->{$in} and 'ARRAY' eq ref $deltas->{$in} ? @{$deltas->{$in}} : ($deltas->{$in}));
            $stmt .= ' AND ' . $db->columnidentifier('delta') . ' ' . $in . ' (' . substr(',?' x scalar @values,1) . ')';
            push(@params,@values);
        }
    } elsif (defined $deltas and length($deltas) > 0) {
        $stmt .= ' AND ' . $db->columnidentifier('delta') . ' = ?';
        push(@params,$deltas);
    }

    return $db->db_get_value($stmt,@params);

}

sub buildrecords_fromrows {

    my ($rows,$load_recursive) = @_;

    my @records = ();
    my $record;

    if (defined $rows and ref $rows eq 'ARRAY') {
        foreach my $row (@$rows) {
            $record = __PACKAGE__->new($row);

            # transformations go here ...

            push @records,$record;
        }
    }

    return \@records;

}

sub process_records {

    my %params = @_;
    my ($process_code,
        $static_context,
        $init_process_context_code,
        $uninit_process_context_code,
        $multithreading,
        $numofthreads,
        $load_recursive) = @params{qw/
            process_code
            static_context
            init_process_context_code
            uninit_process_context_code
            multithreading
            numofthreads
            load_recursive
        /};

    check_table();

    return process_table(
        get_db                      => $get_db,
        class                       => __PACKAGE__,
        process_code                => sub {
                my ($context,$rowblock,$row_offset) = @_;
                return &$process_code($context,buildrecords_fromrows($rowblock,$load_recursive),$row_offset);
            },
        static_context              => $static_context,
        init_process_context_code   => $init_process_context_code,
        uninit_process_context_code => $uninit_process_context_code,
        destroy_reader_dbs_code     => \&destroy_all_dbs,
        multithreading              => $multithreading,
        tableprocessing_threads     => $numofthreads,
    );
}

sub getinsertstatement {

    my ($insert_ignore) = @_;
    check_table();
    return insert_stmt($get_db,__PACKAGE__,$insert_ignore);

}

sub getupsertstatement {

    check_table();
    my $db = &$get_db();
    my $table = $db->tableidentifier($tablename);
    my $upsert_stmt = 'INSERT OR REPLACE INTO ' . $table . ' (' .
      join(', ',map { local $_ = $_; $_ = $db->columnidentifier($_); $_; } @$expected_fieldnames) . ')';
    my @values = ();
    foreach my $fieldname (@$expected_fieldnames) {
        if ('delta' eq $fieldname) {
            my $stmt = 'SELECT \'' . $updated_delta . '\' FROM ' . $table . ' WHERE ' .
                $db->columnidentifier('lrn_code') . ' = ?' .
                ' AND ' . $db->columnidentifier('ported_number') . ' = ?';
            push(@values,'COALESCE((' . $stmt . '), \'' . $added_delta . '\')');
        } else {
            push(@values,'?');
        }
    }
    $upsert_stmt .= ' VALUES (' . join(',',@values) . ')';
    return $upsert_stmt;

}

sub gettablename {

    return $tablename;

}

sub check_table {

    return checktableinfo($get_db,
                   __PACKAGE__,$tablename,
                   $expected_fieldnames,
                   $indexes);

}

1;
